/*
;======================================================================
;HCA - Helicopter Collision Avoidance
;ArmA Version
;by Mr.Peanut
;Version 1.1 : July 1, 2008
;Formerly called HeliCAS, before I knew CAS meant Close Air Support.
;
;SYNTAX:  [_heli] ExecVM "smb_hca\hca_arma.sqs"; 
;EXAMPLE: [[H1,H2,H3,H4,H5,H6,H7,H8,...,H?]] ExecVM "smb_hca\hca_arma.sqs"; 
; where H1-H? are the helicopter names. In this case the default values will be used for the input parameters.
;INPUT VARIABLES:
; _heli = array of helicopters 
;
; OR for total control
;
;SYNTAX:  [_heli, _dist, _ spd, _dt, _delay] ExecVM "smb_hca\hca_arma.sqs"; 
;EXAMPLE: [[H1,H2,H3,H4,H5,H6,H7,H8,...,H?],10,2.0,1.0,0.05] ExecVM "smb_hca\hca_arma.sqs"; 
; where H1-H? are the helicopter names
;
;INPUT VARIABLES:
; _heli = array of helicopters 
; _dist = collision detection distance (m)  
; _spd = collision avoidance speed   (m/s)
; _dt = collision prediction time  (seconds)       
; _delay = collision detection delay  (seconds) 
;
;DESCRIPTION:
;  This script works by predicting each helicopter's position _dt seconds from present.  If any of the helicopters are 
;  closer to each other than distance _dist they are nudged apart from each other with a speed based on the value
;  of _spd. The _delay variable is the loop delay between collision checks.  There is a degree of randomness involved.
;  The helicopters will not always successfully avoid collision. In the demo provided there are eight AI helicopters,
;  one of which will crash about 50% of the time in the first thirty seconds.
;
;  I have found the following values to work well for ArmA and provided them as defaults:
;  _dist = 15 
;  _spd = 2.5
;  _dt = 1
;  _delay = 0.05
;
; Increasing _delay above 0.1 is not advised if you have more than three helicopters. Increasing _spd above 5 
; will make the helicopters flit and bounce around unnaturally.  If you  decrease _dist,it is a good idea to increase _dt.
; For fewer helicopters, decrease _dist, decrease _spd, decrease _dt and increase _delay.
;
;INSTALLATION:
; Copy folder smb_hca into your mission folder.
;
;WARNINGS:
;  1)There is no check for whether the array _hlist is a valid list of helicopters.  I did not want to
;    hard code a list of helicopter types.
;  2)Do not run multiple instances of this script simultaneously.
;  3)Helicopters must be flying at an altitude greater than 0.5m
;
;MP USAGE:
;  HCA will work in multiplayer if is run with the same arguments on all nodes i.e. server and clients.
;  It will also work with helicopters piloted by players, and AI helicopters under player command.
;  There are limitations. Two helicopters approaching collision at high speed might crash. A player trying very hard
;  to crash his helicopter into another will eventually succeed.
;
;  If the helicopters are all AI and will be local to the server, you need only run the script on the server.
;  That was the original usage of this script.
;
;======================================================================
;
; OFPEC Keywords: helicas,helicopter,helicopters,heli,helis,chopper,choppers,collision,collide,avoid,avoidance,crash
; SMUC Code: M for Multiplayer
; Originally released on www.ofpec.com
;
; Thanks to Mandoble for ArmA conversion advice and prodding me to re-release.
;
;Changelog
;2008/06/10 - Optimised code for sqs syntax. Added prefix tags to all global variables.
;             Tested for MP compatibility including player and player controlled pilots.
;             Functions no longer need to be initialised in init.sqs
;2008/07/01 - ArmA conversion. Tweaked default values for ArmA.
*/

if (count _this == 0) exitWith {hint "Error in script HCA. No arguments provided. Exiting script.";};
if (isNil "smb_hca_Dist3D") then {smb_hca_Dist3D = compile preprocessFileLineNumbers "smb_hca\distancepos3d.sqf";};
if (isNil "smb_hca_Sign") then {smb_hca_Sign = compile preprocessFileLineNumbers "smb_hca\sign.sqf";};
if (isNil "smb_hca_Avoid") then {smb_hca_Avoid = compile preprocessFileLineNumbers "smb_hca\avoid_arma.sqf";};

//Default values
_dist = 15;
_spd = 2.5;
_dt = 1;
_delay = 0.05;
//User set values
if ((count _this) == 5) then
{
   _dist = _this select 1;
   _spd = _this select 2;
   _dt = _this select 3 ;
   _delay = _this select 4;
};   
_heli = _this select 0;
//Weed out bad input values
if (_dist <= 0 ) then {_dist = 10;};
if (_spd <= 0) then {_spd = 2;};
if (_dt <= 0 ) then {_dt = 1;};
if (_delay <= 0 ) then {_delay = 0.05;};

//Initialise string list
_hlist = [];

//Main loop
_whileFlag = TRUE;
while {_whileFlag} do
{
   _n = count _heli;
   _i = 0;
   {_hlist set [_i, Format["%1",_i]]; _i = _i + 1 } ForEach _heli;
   //Test for who is alive and update
   _tmpheli = +_heli;
   _tmphlist = +_hlist;
   for [{_i = 0},{_i < _n},{_i = _i + 1}] do
   {
      if (not alive (_heli select _i)) then 
      {
         _tmpheli = _tmpheli - [_heli select _i];
         _tmphlist = _tmphlist - [_hlist select _i];
      };
   };
   _heli = +_tmpheli;
   _hlist = +_tmphlist;
   _tmpheli = NIL;
   _tmphlist = NIL; 
   //hint format ["%1\n%2", _heli,_hlist];
   _n = count _heli;
   //Don't need HCA for only 1 helicopter
   if (_n < 2) exitWith {_whileFlag = FALSE};
   //Acquire present velocities and positions. Predict helicopter positions for _dt seconds
   _newpos = [];
   _hvel = [];
   for [{_i = 0},{_i < _n},{_i = _i + 1}] do
   {
      _vel = velocity (_heli select _i);
      _pos = getPos (_heli select _i);
      _newpos set [_i,[(_pos select 0) + (_vel select 0)*_dt, 
                       (_pos select 1) + (_vel select 1)*_dt,
                       (_pos select 2) + (_vel select 2)*_dt]];
      _hvel set [_i,_vel];
   };
   //Test for possible collisions
   for [{_i = 0},{_i < _n-1},{_i = _i + 1}] do
   {
      for [{_j = _i + 1},{_j < _n},{_j = _j + 1}] do
      {
         if (([_newpos select _i, _newpos select _j] call smb_hca_Dist3D) < _dist and 
             alive ( _heli select _i) and alive (_heli select _j)) then
         {
            [_heli select _i,_heli select _j,_newpos select _i,_newpos select _j,
             _hvel select _i,_hvel select _j,_hlist Select _i,_hlist Select _j,_spd] spawn smb_hca_Avoid;
         
         };        
     };
   };
   sleep _delay;
//Locality check. Sleep if no helis local.
   while {{local _x} count _heli < 1} do
   {
      sleep 5;
   };
};

